from cl_gym.backbones.base import ContinualBackbone
from cl_gym.backbones.mlp import MLP2Layers
from cl_gym.backbones.cnn import CNN1D
from cl_gym.backbones.resnet import ResNet18Small

__all__ = ['ContinualBackbone', 'MLP2Layers', 'CNN1D', 'ResNet18Small']
